@@ -1,7 +1,7 @@
 /*
  * wee-network.c - network functions
  *
- * Copyright (C) 2003-2015 Sébastien Helleu <flashcode@flashtux.org>
+ * Copyright (C) 2003-2021 Sébastien Helleu <flashcode@flashtux.org>
  * Copyright (C) 2005-2010 Emmanuel Bouthenot <kolter@openics.org>
  * Copyright (C) 2010 Gu1ll4um3r0m41n <aeroxteam@gmail.com>
  * Copyright (C) 2012 Simon Arlott
@@ -19,7 +19,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with WeeChat.  If not, see <http://www.gnu.org/licenses/>.
+ * along with WeeChat.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -42,7 +42,9 @@
 #include <poll.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#include <arpa/nameser.h>
 #include <netdb.h>
+#include <resolv.h>
 #include <errno.h>
 #include <gcrypt.h>
 #include <sys/time.h>
@@ -50,25 +52,25 @@
 #include <sys/uio.h>
 #endif
 
-#ifdef HAVE_GNUTLS
 #include <gnutls/gnutls.h>
-#endif
 
 #include "weechat.h"
 #include "wee-network.h"
+#include "wee-config.h"
 #include "wee-eval.h"
+#include "wee-hashtable.h"
 #include "wee-hook.h"
 #include "wee-config.h"
 #include "wee-proxy.h"
 #include "wee-string.h"
+#include "../gui/gui-chat.h"
 #include "../plugins/plugin.h"
 
 
 int network_init_gnutls_ok = 0;
+int network_num_certs = 0;        /* number of certs loaded (system + user) */
 
-#ifdef HAVE_GNUTLS
 gnutls_certificate_credentials_t gnutls_xcred; /* GnuTLS client credentials */
-#endif /* HAVE_GNUTLS */
 
 
 /*
@@ -78,40 +80,180 @@ gnutls_certificate_credentials_t gnutls_xcred; /* GnuTLS client credentials */
 void
 network_init_gcrypt ()
 {
-    if (!weechat_no_gcrypt)
-    {
-        gcry_check_version (GCRYPT_VERSION);
-        gcry_control (GCRYCTL_DISABLE_SECMEM, 0);
-        gcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);
-    }
+    if (weechat_no_gcrypt)
+        return;
+
+    gcry_check_version (GCRYPT_VERSION);
+    gcry_control (GCRYCTL_DISABLE_SECMEM, 0);
+    gcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);
 }
 
 /*
- * Sets trust file with option "gnutls_ca_file".
+ * Loads system's default trusted certificate authorities.
+ *
+ * Returns the number of certificates loaded.
+ */
+
+int
+network_load_system_ca_file (int force_display)
+{
+    int rc;
+
+    if (weechat_no_gnutls)
+        return 0;
+
+    if (!CONFIG_BOOLEAN(config_network_gnutls_ca_system))
+        return 0;
+
+    rc = gnutls_certificate_set_x509_system_trust (gnutls_xcred);
+    if (rc < 0)
+    {
+        gui_chat_printf (
+            NULL,
+            _("%sWarning: failed to load system certificate authorities"),
+            gui_chat_prefix[GUI_CHAT_PREFIX_ERROR]);
+        return 0;
+    }
+
+    if (force_display || (weechat_debug_core >= 1))
+    {
+        gui_chat_printf (
+            NULL,
+            NG_("%d certificate loaded (system)",
+                "%d certificates loaded (system)",
+                rc),
+            rc);
+    }
+
+    return rc;
+}
+
+/*
+ * Loads user's trusted certificate authorities.
+ */
+
+int
+network_load_user_ca_files (int force_display)
+{
+    int i, rc, num_loaded, num_paths;
+    char **paths, *ca_path;
+    struct t_hashtable *options;
+
+    if (weechat_no_gnutls)
+        return 0;
+
+    num_loaded = 0;
+    paths = NULL;
+    options = NULL;
+
+    paths = string_split (CONFIG_STRING(config_network_gnutls_ca_user), ":",
+                          NULL, 0, 0, &num_paths);
+    if (!paths)
+        goto end;
+
+    options = hashtable_new (
+        32,
+        WEECHAT_HASHTABLE_STRING,
+        WEECHAT_HASHTABLE_STRING,
+        NULL, NULL);
+    if (!options)
+        goto end;
+    hashtable_set (options, "directory", "config");
+
+    for (i = 0; i < num_paths; i++)
+    {
+        ca_path = string_eval_path_home (paths[i], NULL, NULL, options);
+        if (ca_path && ca_path[0])
+        {
+            if (access (ca_path, R_OK) == 0)
+            {
+                rc = gnutls_certificate_set_x509_trust_file (
+                    gnutls_xcred, ca_path, GNUTLS_X509_FMT_PEM);
+                if (rc < 0)
+                {
+                    gui_chat_printf (
+                        NULL,
+                        _("%sWarning: failed to load certificate authorities "
+                          "from file %s"),
+                        gui_chat_prefix[GUI_CHAT_PREFIX_ERROR],
+                        ca_path);
+                }
+                else
+                {
+                    num_loaded += rc;
+                    if (force_display || (weechat_debug_core >= 1))
+                    {
+                        gui_chat_printf (
+                            NULL,
+                            NG_("%d certificate loaded (file: %s)",
+                                "%d certificates loaded (file: %s)",
+                                rc),
+                            rc, ca_path);
+                    }
+                }
+            }
+            else
+            {
+                gui_chat_printf (
+                    NULL,
+                    _("%sWarning: failed to load certificate authorities "
+                      "from file %s (file not found)"),
+                    gui_chat_prefix[GUI_CHAT_PREFIX_ERROR],
+                    ca_path);
+            }
+        }
+        if (ca_path)
+            free (ca_path);
+    }
+
+end:
+    if (paths)
+        string_free_split (paths);
+    if (options)
+        hashtable_free (options);
+
+    return num_loaded;
+}
+
+/*
+ * Loads system's default and user's trusted certificate authorities.
  */
 
 void
-network_set_gnutls_ca_file ()
+network_load_ca_files (int force_display)
 {
-#ifdef HAVE_GNUTLS
-    char *ca_path, *ca_path2;
-
     if (weechat_no_gnutls)
         return;
 
-    ca_path = string_expand_home (CONFIG_STRING(config_network_gnutls_ca_file));
-    if (ca_path)
+    network_num_certs = 0;
+
+    network_num_certs += network_load_system_ca_file (force_display);
+    network_num_certs += network_load_user_ca_files (force_display);
+}
+
+/*
+ * Reloads system's default and user's trusted certificate authorities.
+ */
+
+void
+network_reload_ca_files (int force_display)
+{
+    if (weechat_no_gnutls)
+        return;
+
+    gnutls_certificate_free_credentials (gnutls_xcred);
+    if (force_display || (weechat_debug_core >= 1))
     {
-        ca_path2 = string_replace (ca_path, "%h", weechat_home);
-        if (ca_path2)
-        {
-            gnutls_certificate_set_x509_trust_file (gnutls_xcred, ca_path2,
-                                                    GNUTLS_X509_FMT_PEM);
-            free (ca_path2);
-        }
-        free (ca_path);
+        gui_chat_printf (NULL,
+                         NG_("%d certificate purged",
+                             "%d certificates purged",
+                             network_num_certs),
+                         network_num_certs);
     }
-#endif /* HAVE_GNUTLS */
+
+    gnutls_certificate_allocate_credentials (&gnutls_xcred);
+
+    network_load_ca_files (force_display);
 }
 
 /*
@@ -121,13 +263,11 @@ network_set_gnutls_ca_file ()
 void
 network_init_gnutls ()
 {
-#ifdef HAVE_GNUTLS
     if (!weechat_no_gnutls)
     {
         gnutls_global_init ();
         gnutls_certificate_allocate_credentials (&gnutls_xcred);
-
-        network_set_gnutls_ca_file ();
+        network_load_ca_files (0);
 #if LIBGNUTLS_VERSION_NUMBER >= 0x02090a /* 2.9.10 */
         gnutls_certificate_set_verify_function (gnutls_xcred,
                                                 &hook_connect_gnutls_verify_certificates);
@@ -140,7 +280,6 @@ network_init_gnutls ()
                                                          &hook_connect_gnutls_set_certificates);
 #endif /* LIBGNUTLS_VERSION_NUMBER >= 0x020b00 */
     }
-#endif /* HAVE_GNUTLS */
 
     network_init_gnutls_ok = 1;
 }
@@ -154,17 +293,44 @@ network_end ()
 {
     if (network_init_gnutls_ok)
     {
-#ifdef HAVE_GNUTLS
         if (!weechat_no_gnutls)
         {
             gnutls_certificate_free_credentials (gnutls_xcred);
-            gnutls_global_deinit();
+            gnutls_global_deinit ();
         }
-#endif /* HAVE_GNUTLS */
         network_init_gnutls_ok = 0;
     }
 }
 
+/*
+ * Checks if a string contains a valid IP address (IPv4 or IPv6).
+ *
+ * Returns:
+ *   1: string is a valid IPv4 or IPv6
+ *   0: string is not a valid IP address
+ */
+
+int
+network_is_ip_address (const char *address)
+{
+    struct sockaddr_in server_addr;
+    struct sockaddr_in6 server_addr6;
+
+    if (!address || !address[0])
+        return 0;
+
+    /* valid IPv4? */
+    if (inet_pton (AF_INET, address, &server_addr.sin_addr))
+        return 1;
+
+    /* valid IPv6? */
+    if (inet_pton (AF_INET6, address, &server_addr6.sin6_addr))
+        return 1;
+
+    /* not a valid IP address */
+    return 0;
+}
+
 /*
  * Sends data on a socket with retry.
  *
@@ -244,7 +410,7 @@ int
 network_pass_httpproxy (struct t_proxy *proxy, int sock, const char *address,
                         int port)
 {
-    char buffer[256], authbuf[128], authbuf_base64[512], *username, *password;
+    char buffer[4096], authbuf[128], authbuf_base64[512], *username, *password;
     int length;
 
     if (CONFIG_STRING(proxy->options[PROXY_OPTION_USERNAME])
@@ -265,7 +431,8 @@ network_pass_httpproxy (struct t_proxy *proxy, int sock, const char *address,
         snprintf (authbuf, sizeof (authbuf), "%s:%s", username, password);
         free (username);
         free (password);
-        string_encode_base64 (authbuf, strlen (authbuf), authbuf_base64);
+        if (string_base64_encode (authbuf, strlen (authbuf), authbuf_base64) < 0)
+            return 0;
         length = snprintf (buffer, sizeof (buffer),
                            "CONNECT %s:%d HTTP/1.0\r\nProxy-Authorization: "
                            "Basic %s\r\n\r\n",
@@ -311,13 +478,15 @@ network_resolve (const char *hostname, char *ip, int *version)
 
     res = NULL;
 
+    res_init ();
+
     if (getaddrinfo (hostname, NULL, NULL, &res) != 0)
         return 0;
 
     if (!res)
         return 0;
 
-    if (getnameinfo (res->ai_addr, res->ai_addrlen, ipbuffer, sizeof(ipbuffer),
+    if (getnameinfo (res->ai_addr, res->ai_addrlen, ipbuffer, sizeof (ipbuffer),
                      NULL, 0, NI_NUMERICHOST) != 0)
     {
         freeaddrinfo (res);
@@ -340,7 +509,7 @@ network_resolve (const char *hostname, char *ip, int *version)
 /*
  * Establishes a connection and authenticates with a socks4 proxy.
  *
- * The socks4 protocol is explained here: http://en.wikipedia.org/wiki/SOCKS
+ * The socks4 protocol is explained here: https://en.wikipedia.org/wiki/SOCKS
  *
  * WARNING: this function is blocking, it must be called only in a forked
  * process.
@@ -490,9 +659,9 @@ network_pass_socks5proxy (struct t_proxy *proxy, int sock, const char *address,
     }
 
     /* authentication successful then giving address/port to connect */
-    addr_len = strlen(address);
+    addr_len = strlen (address);
     addr_buffer_len = 4 + 1 + addr_len + 2;
-    addr_buffer = malloc (addr_buffer_len * sizeof(*addr_buffer));
+    addr_buffer = malloc (addr_buffer_len * sizeof (*addr_buffer));
     if (!addr_buffer)
         return 0;
     addr_buffer[0] = 5;   /* version 5 */
@@ -670,6 +839,9 @@ network_connect_to (const char *proxy, struct sockaddr *address,
     sock = -1;
     proxy_addrinfo = NULL;
 
+    if (!address || (address_length == 0))
+        return -1;
+
     ptr_proxy = NULL;
     if (proxy && proxy[0])
     {
@@ -696,6 +868,7 @@ network_connect_to (const char *proxy, struct sockaddr *address,
         hints.ai_flags = AI_NUMERICSERV;
         snprintf (str_port, sizeof (str_port), "%d",
                  CONFIG_INTEGER(ptr_proxy->options[PROXY_OPTION_PORT]));
+        res_init ();
         if (getaddrinfo (CONFIG_STRING(ptr_proxy->options[PROXY_OPTION_ADDRESS]),
                          str_port, &hints, &proxy_addrinfo) != 0)
         {
@@ -749,16 +922,12 @@ network_connect_child (struct t_hook *hook_connect)
     char status_without_string[1 + 5 + 1];
     const char *error;
     int rc, length, num_written;
-    int sock, set, flags;
-#ifdef HOOK_CONNECT_MAX_SOCKETS
-    int j;
-#else
+    int sock, set, flags, j;
     struct msghdr msg;
     struct cmsghdr *cmsg;
     char msg_buf[CMSG_SPACE(sizeof (sock))];
     struct iovec iov[1];
     char iov_data[1] = { 0 };
-#endif /* HOOK_CONNECT_MAX_SOCKETS */
     /*
      * indicates that something is wrong with whichever group of
      * servers is being tried first after connecting, so start at
@@ -806,6 +975,7 @@ network_connect_child (struct t_hook *hook_connect)
 #ifdef AI_ADDRCONFIG
     hints.ai_flags = AI_ADDRCONFIG;
 #endif /* AI_ADDRCONFIG */
+    res_init ();
     if (ptr_proxy)
     {
         hints.ai_family = (CONFIG_BOOLEAN(ptr_proxy->options[PROXY_OPTION_IPV6])) ?
@@ -979,7 +1149,7 @@ network_connect_child (struct t_hook *hook_connect)
             if (tmp_num_groups >= retry)
             {
                 /* shuffle while adding */
-                rand_num = tmp_host + (rand() % ((i + 1) - tmp_host));
+                rand_num = tmp_host + (rand () % ((i + 1) - tmp_host));
                 if (rand_num == i)
                     res_reorder[i++] = ptr_res;
                 else
@@ -1010,7 +1180,7 @@ network_connect_child (struct t_hook *hook_connect)
             if (tmp_num_groups < retry)
             {
                 /* shuffle while adding */
-                rand_num = tmp_host + (rand() % ((i + 1) - tmp_host));
+                rand_num = tmp_host + (rand () % ((i + 1) - tmp_host));
                 if (rand_num == i)
                     res_reorder[i++] = ptr_res;
                 else
@@ -1044,38 +1214,41 @@ network_connect_child (struct t_hook *hook_connect)
     {
         ptr_res = res_reorder[i];
 
-#ifdef HOOK_CONNECT_MAX_SOCKETS
-        /* use pre-created socket pool */
-        sock = -1;
-        for (j = 0; j < HOOK_CONNECT_MAX_SOCKETS; j++)
+        if (hook_socketpair_ok)
         {
-            if (ptr_res->ai_family == AF_INET)
+            /* create a socket */
+            sock = socket (ptr_res->ai_family,
+                           ptr_res->ai_socktype,
+                           ptr_res->ai_protocol);
+        }
+        else
+        {
+            /* use pre-created socket pool */
+            sock = -1;
+            for (j = 0; j < HOOK_CONNECT_MAX_SOCKETS; j++)
             {
-                sock = HOOK_CONNECT(hook_connect, sock_v4[j]);
-                if (sock != -1)
+                if (ptr_res->ai_family == AF_INET)
                 {
-                    HOOK_CONNECT(hook_connect, sock_v4[j]) = -1;
-                    break;
+                    sock = HOOK_CONNECT(hook_connect, sock_v4[j]);
+                    if (sock != -1)
+                    {
+                        HOOK_CONNECT(hook_connect, sock_v4[j]) = -1;
+                        break;
+                    }
                 }
-            }
-            else if (ptr_res->ai_family == AF_INET6)
-            {
-                sock = HOOK_CONNECT(hook_connect, sock_v6[j]);
-                if (sock != -1)
+                else if (ptr_res->ai_family == AF_INET6)
                 {
-                    HOOK_CONNECT(hook_connect, sock_v6[j]) = -1;
-                    break;
+                    sock = HOOK_CONNECT(hook_connect, sock_v6[j]);
+                    if (sock != -1)
+                    {
+                        HOOK_CONNECT(hook_connect, sock_v6[j]) = -1;
+                        break;
+                    }
                 }
             }
+            if (sock < 0)
+                continue;
         }
-        if (sock < 0)
-            continue;
-#else
-        /* create a socket */
-        sock = socket (ptr_res->ai_family,
-                       ptr_res->ai_socktype,
-                       ptr_res->ai_protocol);
-#endif /* HOOK_CONNECT_MAX_SOCKETS */
         if (sock < 0)
         {
             status_str[0] = '0' + WEECHAT_HOOK_CONNECT_SOCKET_ERROR;
@@ -1183,30 +1356,36 @@ network_connect_child (struct t_hook *hook_connect)
         }
 
         /* send the socket to the parent process */
-#ifndef HOOK_CONNECT_MAX_SOCKETS
-        memset (&msg, 0, sizeof (msg));
-        msg.msg_control = msg_buf;
-        msg.msg_controllen = sizeof (msg_buf);
+        if (hook_socketpair_ok)
+        {
+            memset (&msg, 0, sizeof (msg));
+            msg.msg_control = msg_buf;
+            msg.msg_controllen = sizeof (msg_buf);
 
-        /* send 1 byte of data (not required on Linux, required by BSD/OSX) */
-        memset (iov, 0, sizeof (iov));
-        iov[0].iov_base = iov_data;
-        iov[0].iov_len = 1;
-        msg.msg_iov = iov;
-        msg.msg_iovlen = 1;
+            /*
+             * send 1 byte of data
+             * (not required on Linux, required by BSD/macOS)
+             */
+            memset (iov, 0, sizeof (iov));
+            iov[0].iov_base = iov_data;
+            iov[0].iov_len = 1;
+            msg.msg_iov = iov;
+            msg.msg_iovlen = 1;
 
-        cmsg = CMSG_FIRSTHDR(&msg);
-        cmsg->cmsg_level = SOL_SOCKET;
-        cmsg->cmsg_type = SCM_RIGHTS;
-        cmsg->cmsg_len = CMSG_LEN(sizeof (sock));
-        memcpy(CMSG_DATA(cmsg), &sock, sizeof (sock));
-        msg.msg_controllen = cmsg->cmsg_len;
-        num_written = sendmsg (HOOK_CONNECT(hook_connect, child_send), &msg, 0);
-        (void) num_written;
-#else
-        num_written = write (HOOK_CONNECT(hook_connect, child_write), &sock, sizeof (sock));
-        (void) num_written;
-#endif /* HOOK_CONNECT_MAX_SOCKETS */
+            cmsg = CMSG_FIRSTHDR(&msg);
+            cmsg->cmsg_level = SOL_SOCKET;
+            cmsg->cmsg_type = SCM_RIGHTS;
+            cmsg->cmsg_len = CMSG_LEN(sizeof (sock));
+            memcpy (CMSG_DATA(cmsg), &sock, sizeof (sock));
+            msg.msg_controllen = cmsg->cmsg_len;
+            num_written = sendmsg (HOOK_CONNECT(hook_connect, child_send), &msg, 0);
+            (void) num_written;
+        }
+        else
+        {
+            num_written = write (HOOK_CONNECT(hook_connect, child_write), &sock, sizeof (sock));
+            (void) num_written;
+        }
     }
     else
     {
@@ -1233,19 +1412,22 @@ end:
  */
 
 int
-network_connect_child_timer_cb (void *arg_hook_connect, int remaining_calls)
+network_connect_child_timer_cb (const void *pointer, void *data,
+                                int remaining_calls)
 {
     struct t_hook *hook_connect;
 
     /* make C compiler happy */
+    (void) data;
     (void) remaining_calls;
 
-    hook_connect = (struct t_hook *)arg_hook_connect;
+    hook_connect = (struct t_hook *)pointer;
 
     HOOK_CONNECT(hook_connect, hook_child_timer) = NULL;
 
     (void) (HOOK_CONNECT(hook_connect, callback))
-        (hook_connect->callback_data,
+        (hook_connect->callback_pointer,
+         hook_connect->callback_data,
          WEECHAT_HOOK_CONNECT_TIMEOUT,
          0, -1, NULL, NULL);
     unhook (hook_connect);
@@ -1260,17 +1442,18 @@ network_connect_child_timer_cb (void *arg_hook_connect, int remaining_calls)
  * finish).
  */
 
-#ifdef HAVE_GNUTLS
 int
-network_connect_gnutls_handshake_fd_cb (void *arg_hook_connect, int fd)
+network_connect_gnutls_handshake_fd_cb (const void *pointer, void *data,
+                                        int fd)
 {
     struct t_hook *hook_connect;
     int rc, direction, flags;
 
     /* make C compiler happy */
+    (void) data;
     (void) fd;
 
-    hook_connect = (struct t_hook *)arg_hook_connect;
+    hook_connect = (struct t_hook *)pointer;
 
     rc = gnutls_handshake (*HOOK_CONNECT(hook_connect, gnutls_sess));
 
@@ -1289,8 +1472,10 @@ network_connect_gnutls_handshake_fd_cb (void *arg_hook_connect, int fd)
     }
     else if (rc != GNUTLS_E_SUCCESS)
     {
+        unhook (HOOK_CONNECT(hook_connect, handshake_hook_fd));
         (void) (HOOK_CONNECT(hook_connect, callback))
-            (hook_connect->callback_data,
+            (hook_connect->callback_pointer,
+             hook_connect->callback_data,
              WEECHAT_HOOK_CONNECT_GNUTLS_HANDSHAKE_ERROR, rc,
              HOOK_CONNECT(hook_connect, sock),
              gnutls_strerror (rc),
@@ -1309,8 +1494,10 @@ network_connect_gnutls_handshake_fd_cb (void *arg_hook_connect, int fd)
          */
         if (hook_connect_gnutls_verify_certificates (*HOOK_CONNECT(hook_connect, gnutls_sess)) != 0)
         {
+            unhook (HOOK_CONNECT(hook_connect, handshake_hook_fd));
             (void) (HOOK_CONNECT(hook_connect, callback))
-                (hook_connect->callback_data,
+                (hook_connect->callback_pointer,
+                 hook_connect->callback_data,
                  WEECHAT_HOOK_CONNECT_GNUTLS_HANDSHAKE_ERROR, rc,
                  HOOK_CONNECT(hook_connect, sock),
                  "Error in the certificate.",
@@ -1321,77 +1508,74 @@ network_connect_gnutls_handshake_fd_cb (void *arg_hook_connect, int fd)
 #endif /* LIBGNUTLS_VERSION_NUMBER < 0x02090a */
         unhook (HOOK_CONNECT(hook_connect, handshake_hook_fd));
         (void) (HOOK_CONNECT(hook_connect, callback))
-                (hook_connect->callback_data,
-                 WEECHAT_HOOK_CONNECT_OK, 0,
-                 HOOK_CONNECT(hook_connect, sock),
-                 NULL, HOOK_CONNECT(hook_connect, handshake_ip_address));
+            (hook_connect->callback_pointer,
+             hook_connect->callback_data,
+             WEECHAT_HOOK_CONNECT_OK, 0,
+             HOOK_CONNECT(hook_connect, sock),
+             NULL, HOOK_CONNECT(hook_connect, handshake_ip_address));
         unhook (hook_connect);
     }
 
     return WEECHAT_RC_OK;
 }
-#endif /* HAVE_GNUTLS */
 
 /*
  * Timer callback for timeout of handshake.
  */
 
-#ifdef HAVE_GNUTLS
 int
-network_connect_gnutls_handshake_timer_cb (void *arg_hook_connect,
+network_connect_gnutls_handshake_timer_cb (const void *pointer,
+                                           void *data,
                                            int remaining_calls)
 {
     struct t_hook *hook_connect;
 
     /* make C compiler happy */
+    (void) data;
     (void) remaining_calls;
 
-    hook_connect = (struct t_hook *)arg_hook_connect;
+    hook_connect = (struct t_hook *)pointer;
 
     HOOK_CONNECT(hook_connect, handshake_hook_timer) = NULL;
 
+    unhook (HOOK_CONNECT(hook_connect, handshake_hook_fd));
     (void) (HOOK_CONNECT(hook_connect, callback))
-            (hook_connect->callback_data,
-             WEECHAT_HOOK_CONNECT_GNUTLS_HANDSHAKE_ERROR,
-             GNUTLS_E_EXPIRED,
-             HOOK_CONNECT(hook_connect, sock),
-             gnutls_strerror (GNUTLS_E_EXPIRED),
-             HOOK_CONNECT(hook_connect, handshake_ip_address));
+        (hook_connect->callback_pointer,
+         hook_connect->callback_data,
+         WEECHAT_HOOK_CONNECT_GNUTLS_HANDSHAKE_ERROR,
+         GNUTLS_E_EXPIRED,
+         HOOK_CONNECT(hook_connect, sock),
+         gnutls_strerror (GNUTLS_E_EXPIRED),
+         HOOK_CONNECT(hook_connect, handshake_ip_address));
     unhook (hook_connect);
 
     return WEECHAT_RC_OK;
 }
-#endif /* HAVE_GNUTLS */
 
 /*
  * Reads connection progress from child process.
  */
 
 int
-network_connect_child_read_cb (void *arg_hook_connect, int fd)
+network_connect_child_read_cb (const void *pointer, void *data, int fd)
 {
     struct t_hook *hook_connect;
     char buffer[1], buf_size[6], *cb_error, *cb_ip_address, *error;
     int num_read;
     long size_msg;
-#ifdef HAVE_GNUTLS
     int rc, direction;
-#endif /* HAVE_GNUTLS */
-    int sock;
-#ifdef HOOK_CONNECT_MAX_SOCKETS
-    int i;
-#else
+    int sock, i;
     struct msghdr msg;
     struct cmsghdr *cmsg;
     char msg_buf[CMSG_SPACE(sizeof (sock))];
     struct iovec iov[1];
     char iov_data[1];
-#endif /* HOOK_CONNECT_MAX_SOCKETS */
 
     /* make C compiler happy */
+    (void) data;
     (void) fd;
 
-    hook_connect = (struct t_hook *)arg_hook_connect;
+    hook_connect = (struct t_hook *)pointer;
 
     cb_error = NULL;
     cb_ip_address = NULL;
@@ -1429,48 +1613,53 @@ network_connect_child_read_cb (void *arg_hook_connect, int fd)
                 }
             }
 
-#ifndef HOOK_CONNECT_MAX_SOCKETS
-            /* receive the socket from the child process */
-            memset (&msg, 0, sizeof (msg));
-            msg.msg_control = msg_buf;
-            msg.msg_controllen = sizeof (msg_buf);
-
-            /* recv 1 byte of data (not required on Linux, required by BSD/OSX) */
-            memset (iov, 0, sizeof (iov));
-            iov[0].iov_base = iov_data;
-            iov[0].iov_len = 1;
-            msg.msg_iov = iov;
-            msg.msg_iovlen = 1;
-
-            if (recvmsg (HOOK_CONNECT(hook_connect, child_recv), &msg, 0) >= 0)
+            if (hook_socketpair_ok)
             {
-                cmsg = CMSG_FIRSTHDR(&msg);
-                if (cmsg != NULL
-                    && cmsg->cmsg_level == SOL_SOCKET
-                    && cmsg->cmsg_type == SCM_RIGHTS
-                    && cmsg->cmsg_len >= sizeof (sock))
+                /* receive the socket from the child process */
+                memset (&msg, 0, sizeof (msg));
+                msg.msg_control = msg_buf;
+                msg.msg_controllen = sizeof (msg_buf);
+
+                /*
+                 * recv 1 byte of data
+                 * (not required on Linux, required by BSD/macOS)
+                 */
+                memset (iov, 0, sizeof (iov));
+                iov[0].iov_base = iov_data;
+                iov[0].iov_len = 1;
+                msg.msg_iov = iov;
+                msg.msg_iovlen = 1;
+
+                if (recvmsg (HOOK_CONNECT(hook_connect, child_recv), &msg, 0) >= 0)
                 {
-                    memcpy(&sock, CMSG_DATA(cmsg), sizeof (sock));
+                    cmsg = CMSG_FIRSTHDR(&msg);
+                    if (cmsg != NULL
+                        && cmsg->cmsg_level == SOL_SOCKET
+                        && cmsg->cmsg_type == SCM_RIGHTS
+                        && cmsg->cmsg_len >= sizeof (sock))
+                    {
+                        memcpy (&sock, CMSG_DATA(cmsg), sizeof (sock));
+                    }
                 }
             }
-#else
-            num_read = read (HOOK_CONNECT(hook_connect, child_read), &sock, sizeof (sock));
-            (void) num_read;
-
-            /* prevent unhook process from closing the socket */
-            for (i = 0; i < HOOK_CONNECT_MAX_SOCKETS; i++)
+            else
             {
-                if (HOOK_CONNECT(hook_connect, sock_v4[i]) == sock)
-                    HOOK_CONNECT(hook_connect, sock_v4[i]) = -1;
+                num_read = read (HOOK_CONNECT(hook_connect, child_read), &sock, sizeof (sock));
+                (void) num_read;
 
-                if (HOOK_CONNECT(hook_connect, sock_v6[i]) == sock)
-                    HOOK_CONNECT(hook_connect, sock_v6[i]) = -1;
+                /* prevent unhook process from closing the socket */
+                for (i = 0; i < HOOK_CONNECT_MAX_SOCKETS; i++)
+                {
+                    if (HOOK_CONNECT(hook_connect, sock_v4[i]) == sock)
+                        HOOK_CONNECT(hook_connect, sock_v4[i]) = -1;
+
+                    if (HOOK_CONNECT(hook_connect, sock_v6[i]) == sock)
+                        HOOK_CONNECT(hook_connect, sock_v6[i]) = -1;
+                }
             }
-#endif /* HOOK_CONNECT_MAX_SOCKETS */
 
             HOOK_CONNECT(hook_connect, sock) = sock;
 
-#ifdef HAVE_GNUTLS
             if (HOOK_CONNECT(hook_connect, gnutls_sess))
             {
                 /*
@@ -1507,19 +1696,20 @@ network_connect_child_read_cb (void *arg_hook_connect, int fd)
                                  HOOK_CONNECT(hook_connect, sock),
                                  (!direction ? 1 : 0), (direction  ? 1 : 0), 0,
                                  &network_connect_gnutls_handshake_fd_cb,
-                                 hook_connect);
+                                 hook_connect, NULL);
                     HOOK_CONNECT(hook_connect, handshake_hook_timer) =
                         hook_timer (hook_connect->plugin,
                                     CONFIG_INTEGER(config_network_gnutls_handshake_timeout) * 1000,
                                     0, 1,
                                     &network_connect_gnutls_handshake_timer_cb,
-                                    hook_connect);
+                                    hook_connect, NULL);
                     return WEECHAT_RC_OK;
                 }
                 else if (rc != GNUTLS_E_SUCCESS)
                 {
                     (void) (HOOK_CONNECT(hook_connect, callback))
-                        (hook_connect->callback_data,
+                        (hook_connect->callback_pointer,
+                         hook_connect->callback_data,
                          WEECHAT_HOOK_CONNECT_GNUTLS_HANDSHAKE_ERROR,
                          rc, sock,
                          gnutls_strerror (rc),
@@ -1540,7 +1730,8 @@ network_connect_child_read_cb (void *arg_hook_connect, int fd)
                 if (hook_connect_gnutls_verify_certificates (*HOOK_CONNECT(hook_connect, gnutls_sess)) != 0)
                 {
                     (void) (HOOK_CONNECT(hook_connect, callback))
-                        (hook_connect->callback_data,
+                        (hook_connect->callback_pointer,
+                         hook_connect->callback_data,
                          WEECHAT_HOOK_CONNECT_GNUTLS_HANDSHAKE_ERROR,
                          rc, sock,
                          "Error in the certificate.",
@@ -1552,7 +1743,6 @@ network_connect_child_read_cb (void *arg_hook_connect, int fd)
                 }
 #endif /* LIBGNUTLS_VERSION_NUMBER < 0x02090a */
             }
-#endif /* HAVE_GNUTLS */
         }
         else
         {
@@ -1583,15 +1773,19 @@ network_connect_child_read_cb (void *arg_hook_connect, int fd)
             }
         }
         (void) (HOOK_CONNECT(hook_connect, callback))
-            (hook_connect->callback_data, buffer[0] - '0', 0,
+            (hook_connect->callback_pointer,
+             hook_connect->callback_data,
+             buffer[0] - '0', 0,
              sock, cb_error, cb_ip_address);
         unhook (hook_connect);
     }
     else
     {
         (void) (HOOK_CONNECT(hook_connect, callback))
-            (hook_connect->callback_data, WEECHAT_HOOK_CONNECT_MEMORY_ERROR,
-             0, sock, cb_error, cb_ip_address);
+            (hook_connect->callback_pointer,
+             hook_connect->callback_data,
+             WEECHAT_HOOK_CONNECT_MEMORY_ERROR,
+             0, sock, "child_read_cb", NULL);
         unhook (hook_connect);
     }
 
@@ -1610,37 +1804,50 @@ network_connect_child_read_cb (void *arg_hook_connect, int fd)
 void
 network_connect_with_fork (struct t_hook *hook_connect)
 {
-    int child_pipe[2], rc;
-#ifdef HOOK_CONNECT_MAX_SOCKETS
-    int i;
-#else
-    int child_socket[2];
-#endif /* HOOK_CONNECT_MAX_SOCKETS */
-#ifdef HAVE_GNUTLS
+    int child_pipe[2], child_socket[2], rc, i;
+    char str_error[1024];
     const char *pos_error;
-#endif /* HAVE_GNUTLS */
     pid_t pid;
 
-#ifdef HAVE_GNUTLS
     /* initialize GnuTLS if SSL asked */
     if (HOOK_CONNECT(hook_connect, gnutls_sess))
     {
         if (gnutls_init (HOOK_CONNECT(hook_connect, gnutls_sess), GNUTLS_CLIENT) != GNUTLS_E_SUCCESS)
         {
             (void) (HOOK_CONNECT(hook_connect, callback))
-                (hook_connect->callback_data,
+                (hook_connect->callback_pointer,
+                 hook_connect->callback_data,
                  WEECHAT_HOOK_CONNECT_GNUTLS_INIT_ERROR,
                  0, -1, NULL, NULL);
             unhook (hook_connect);
             return;
         }
+        if (!network_is_ip_address (HOOK_CONNECT(hook_connect, address)))
+        {
+            /* set the server name (only if it's NOT an IPv4/IPv6) */
+            rc = gnutls_server_name_set (*HOOK_CONNECT(hook_connect, gnutls_sess),
+                                         GNUTLS_NAME_DNS,
+                                         HOOK_CONNECT(hook_connect, address),
+                                         strlen (HOOK_CONNECT(hook_connect, address)));
+            if (rc != GNUTLS_E_SUCCESS)
+            {
+                (void) (HOOK_CONNECT(hook_connect, callback))
+                    (hook_connect->callback_pointer,
+                     hook_connect->callback_data,
+                     WEECHAT_HOOK_CONNECT_GNUTLS_INIT_ERROR,
+                     0, -1, _("set server name indication (SNI) failed"), NULL);
+                unhook (hook_connect);
+                return;
+            }
+        }
         rc = gnutls_priority_set_direct (*HOOK_CONNECT(hook_connect, gnutls_sess),
                                          HOOK_CONNECT(hook_connect, gnutls_priorities),
                                          &pos_error);
         if (rc != GNUTLS_E_SUCCESS)
         {
             (void) (HOOK_CONNECT(hook_connect, callback))
-                (hook_connect->callback_data,
+                (hook_connect->callback_pointer,
+                 hook_connect->callback_data,
                  WEECHAT_HOOK_CONNECT_GNUTLS_INIT_ERROR,
                  0, -1, _("invalid priorities"), NULL);
             unhook (hook_connect);
@@ -1652,50 +1859,58 @@ network_connect_with_fork (struct t_hook *hook_connect)
         gnutls_transport_set_ptr (*HOOK_CONNECT(hook_connect, gnutls_sess),
                                   (gnutls_transport_ptr_t) ((unsigned long) HOOK_CONNECT(hook_connect, sock)));
     }
-#endif /* HAVE_GNUTLS */
 
     /* create pipe for child process */
     if (pipe (child_pipe) < 0)
     {
         (void) (HOOK_CONNECT(hook_connect, callback))
-            (hook_connect->callback_data,
+            (hook_connect->callback_pointer,
+             hook_connect->callback_data,
              WEECHAT_HOOK_CONNECT_MEMORY_ERROR,
-             0, -1, NULL, NULL);
+             0, -1, "pipe", NULL);
         unhook (hook_connect);
         return;
     }
     HOOK_CONNECT(hook_connect, child_read) = child_pipe[0];
     HOOK_CONNECT(hook_connect, child_write) = child_pipe[1];
 
-#ifndef HOOK_CONNECT_MAX_SOCKETS
-    /* create socket for child process */
-    if (socketpair (AF_LOCAL, SOCK_DGRAM, 0, child_socket) < 0)
+    if (hook_socketpair_ok)
     {
-        (void) (HOOK_CONNECT(hook_connect, callback))
-            (hook_connect->callback_data,
-             WEECHAT_HOOK_CONNECT_MEMORY_ERROR,
-             0, -1, NULL, NULL);
-        unhook (hook_connect);
-        return;
+        /* create socket for child process */
+        if (socketpair (AF_LOCAL, SOCK_DGRAM, 0, child_socket) < 0)
+        {
+            (void) (HOOK_CONNECT(hook_connect, callback))
+                (hook_connect->callback_pointer,
+                 hook_connect->callback_data,
+                 WEECHAT_HOOK_CONNECT_MEMORY_ERROR,
+                 0, -1, "socketpair", NULL);
+            unhook (hook_connect);
+            return;
+        }
+        HOOK_CONNECT(hook_connect, child_recv) = child_socket[0];
+        HOOK_CONNECT(hook_connect, child_send) = child_socket[1];
     }
-    HOOK_CONNECT(hook_connect, child_recv) = child_socket[0];
-    HOOK_CONNECT(hook_connect, child_send) = child_socket[1];
-#else
-    for (i = 0; i < HOOK_CONNECT_MAX_SOCKETS; i++)
+    else
     {
-        HOOK_CONNECT(hook_connect, sock_v4[i]) = socket (AF_INET, SOCK_STREAM, 0);
-        HOOK_CONNECT(hook_connect, sock_v6[i]) = socket (AF_INET6, SOCK_STREAM, 0);
+        for (i = 0; i < HOOK_CONNECT_MAX_SOCKETS; i++)
+        {
+            HOOK_CONNECT(hook_connect, sock_v4[i]) = socket (AF_INET, SOCK_STREAM, 0);
+            HOOK_CONNECT(hook_connect, sock_v6[i]) = socket (AF_INET6, SOCK_STREAM, 0);
+        }
     }
-#endif /* HOOK_CONNECT_MAX_SOCKETS */
 
     switch (pid = fork ())
     {
         /* fork failed */
         case -1:
+            snprintf (str_error, sizeof (str_error),
+                      "fork error: %s",
+                      strerror (errno));
             (void) (HOOK_CONNECT(hook_connect, callback))
-                (hook_connect->callback_data,
+                (hook_connect->callback_pointer,
+                 hook_connect->callback_data,
                  WEECHAT_HOOK_CONNECT_MEMORY_ERROR,
-                 0, -1, NULL, NULL);
+                 0, -1, str_error, NULL);
             unhook (hook_connect);
             return;
         /* child process */
@@ -1703,9 +1918,8 @@ network_connect_with_fork (struct t_hook *hook_connect)
             rc = setuid (getuid ());
             (void) rc;
             close (HOOK_CONNECT(hook_connect, child_read));
-#ifndef HOOK_CONNECT_MAX_SOCKETS
-            close (HOOK_CONNECT(hook_connect, child_recv));
-#endif /* HOOK_CONNECT_MAX_SOCKETS */
+            if (hook_socketpair_ok)
+                close (HOOK_CONNECT(hook_connect, child_recv));
             network_connect_child (hook_connect);
             _exit (EXIT_SUCCESS);
     }
@@ -1713,18 +1927,20 @@ network_connect_with_fork (struct t_hook *hook_connect)
     HOOK_CONNECT(hook_connect, child_pid) = pid;
     close (HOOK_CONNECT(hook_connect, child_write));
     HOOK_CONNECT(hook_connect, child_write) = -1;
-#ifndef HOOK_CONNECT_MAX_SOCKETS
-    close (HOOK_CONNECT(hook_connect, child_send));
-    HOOK_CONNECT(hook_connect, child_send) = -1;
-#endif /* HOOK_CONNECT_MAX_SOCKETS */
+    if (hook_socketpair_ok)
+    {
+        close (HOOK_CONNECT(hook_connect, child_send));
+        HOOK_CONNECT(hook_connect, child_send) = -1;
+    }
     HOOK_CONNECT(hook_connect, hook_child_timer) = hook_timer (hook_connect->plugin,
                                                                CONFIG_INTEGER(config_network_connection_timeout) * 1000,
                                                                0, 1,
                                                                &network_connect_child_timer_cb,
-                                                               hook_connect);
+                                                               hook_connect,
+                                                               NULL);
     HOOK_CONNECT(hook_connect, hook_fd) = hook_fd (hook_connect->plugin,
                                                    HOOK_CONNECT(hook_connect, child_read),
                                                    1, 0, 0,
                                                    &network_connect_child_read_cb,
-                                                   hook_connect);
+                                                   hook_connect, NULL);
 }
